"""Cmd2 testing for argument parsing"""

import argparse

import pytest

import cmd2

from .conftest import (
    run_cmd,
)


class ArgparseApp(cmd2.Cmd):
    def __init__(self) -> None:
        self.maxrepeats = 3
        cmd2.Cmd.__init__(self)

    def namespace_provider(self) -> argparse.Namespace:
        ns = argparse.Namespace()
        ns.custom_stuff = "custom"
        return ns

    @staticmethod
    def _say_parser_builder() -> cmd2.Cmd2ArgumentParser:
        say_parser = cmd2.Cmd2ArgumentParser()
        say_parser.add_argument('-p', '--piglatin', action='store_true', help='atinLay')
        say_parser.add_argument('-s', '--shout', action='store_true', help='N00B EMULATION MODE')
        say_parser.add_argument('-r', '--repeat', type=int, help='output [n] times')
        say_parser.add_argument('words', nargs='+', help='words to say')
        return say_parser

    @cmd2.with_argparser(_say_parser_builder)
    def do_say(self, args, *, keyword_arg: str | None = None) -> None:
        """Repeat what you tell me to.

        :param args: argparse namespace
        :param keyword_arg: Optional keyword arguments
        """
        words = []
        for word in args.words:
            modified_word = word
            if word is None:
                modified_word = ''
            if args.piglatin:
                modified_word = f'{word[1:]}{word[0]}ay'
            if args.shout:
                modified_word = word.upper()
            words.append(modified_word)
        repetitions = args.repeat or 1
        for _ in range(min(repetitions, self.maxrepeats)):
            self.stdout.write(' '.join(words))
            self.stdout.write('\n')

        if keyword_arg is not None:
            print(keyword_arg)

    tag_parser = cmd2.Cmd2ArgumentParser(description='create a html tag')
    tag_parser.add_argument('tag', help='tag')
    tag_parser.add_argument('content', nargs='+', help='content to surround with tag')

    @cmd2.with_argparser(tag_parser, preserve_quotes=True)
    def do_tag(self, args) -> None:
        self.stdout.write('<{0}>{1}</{0}>'.format(args.tag, ' '.join(args.content)))
        self.stdout.write('\n')

    @cmd2.with_argparser(cmd2.Cmd2ArgumentParser(), ns_provider=namespace_provider)
    def do_test_argparse_ns(self, args) -> None:
        self.stdout.write(f'{args.custom_stuff}')

    @cmd2.with_argument_list
    def do_arglist(self, arglist, *, keyword_arg: str | None = None) -> None:
        if isinstance(arglist, list):
            self.stdout.write('True')
        else:
            self.stdout.write('False')

        if keyword_arg is not None:
            print(keyword_arg)

    @cmd2.with_argument_list(preserve_quotes=True)
    def do_preservelist(self, arglist) -> None:
        self.stdout.write(f'{arglist}')

    @classmethod
    def _speak_parser_builder(cls) -> cmd2.Cmd2ArgumentParser:
        known_parser = cmd2.Cmd2ArgumentParser()
        known_parser.add_argument('-p', '--piglatin', action='store_true', help='atinLay')
        known_parser.add_argument('-s', '--shout', action='store_true', help='N00B EMULATION MODE')
        known_parser.add_argument('-r', '--repeat', type=int, help='output [n] times')
        return known_parser

    @cmd2.with_argparser(_speak_parser_builder, with_unknown_args=True)
    def do_speak(self, args, extra, *, keyword_arg: str | None = None) -> None:
        """Repeat what you tell me to."""
        words = []
        for word in extra:
            modified_word = word
            if word is None:
                modified_word = ''
            if args.piglatin:
                modified_word = f'{word[1:]}{word[0]}ay'
            if args.shout:
                modified_word = word.upper()
            words.append(modified_word)
        repetitions = args.repeat or 1
        for _ in range(min(repetitions, self.maxrepeats)):
            self.stdout.write(' '.join(words))
            self.stdout.write('\n')

        if keyword_arg is not None:
            print(keyword_arg)

    @cmd2.with_argparser(cmd2.Cmd2ArgumentParser(), preserve_quotes=True, with_unknown_args=True)
    def do_test_argparse_with_list_quotes(self, args, extra) -> None:
        self.stdout.write('{}'.format(' '.join(extra)))

    @cmd2.with_argparser(cmd2.Cmd2ArgumentParser(), ns_provider=namespace_provider, with_unknown_args=True)
    def do_test_argparse_with_list_ns(self, args, extra) -> None:
        self.stdout.write(f'{args.custom_stuff}')


@pytest.fixture
def argparse_app():
    return ArgparseApp()


def test_invalid_syntax(argparse_app) -> None:
    _out, err = run_cmd(argparse_app, 'speak "')
    assert err[0] == "Invalid syntax: No closing quotation"


def test_argparse_basic_command(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'say hello')
    assert out == ['hello']


def test_argparse_remove_quotes(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'say "hello there"')
    assert out == ['hello there']


def test_argparse_with_no_args(argparse_app) -> None:
    """Make sure we receive TypeError when calling argparse-based function with no args"""
    with pytest.raises(TypeError) as excinfo:
        argparse_app.do_say()
    assert 'Expected arguments' in str(excinfo.value)


def test_argparser_kwargs(argparse_app, capsys) -> None:
    """Test with_argparser wrapper passes through kwargs to command function"""
    argparse_app.do_say('word', keyword_arg="foo")
    out, _err = capsys.readouterr()
    assert out == "foo\n"


def test_argparse_preserve_quotes(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'tag mytag "hello"')
    assert out[0] == '<mytag>"hello"</mytag>'


def test_argparse_custom_namespace(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'test_argparse_ns')
    assert out[0] == 'custom'


def test_argparse_with_list(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'speak -s hello world!')
    assert out == ['HELLO WORLD!']


def test_argparse_with_list_remove_quotes(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'speak -s hello "world!"')
    assert out == ['HELLO WORLD!']


def test_argparse_with_list_preserve_quotes(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'test_argparse_with_list_quotes "hello" person')
    assert out[0] == '"hello" person'


def test_argparse_with_list_custom_namespace(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'test_argparse_with_list_ns')
    assert out[0] == 'custom'


def test_argparse_with_list_and_empty_doc(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'speak -s hello world!')
    assert out == ['HELLO WORLD!']


def test_argparser_correct_args_with_quotes_and_midline_options(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, "speak 'This  is a' -s test of the emergency broadcast system!")
    assert out == ['THIS  IS A TEST OF THE EMERGENCY BROADCAST SYSTEM!']


def test_argparser_and_unknown_args_kwargs(argparse_app, capsys) -> None:
    """Test with_argparser wrapper passing through kwargs to command function"""
    argparse_app.do_speak('', keyword_arg="foo")
    out, _err = capsys.readouterr()
    assert out == "foo\n"


def test_argparse_quoted_arguments_multiple(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'say "hello  there" "rick & morty"')
    assert out == ['hello  there rick & morty']


def test_argparse_help_docstring(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'help say')
    assert out[0].startswith('Usage: say')
    assert out[1] == ''
    assert out[2] == 'Repeat what you tell me to.'
    for line in out:
        assert not line.startswith(':')


def test_argparse_help_description(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'help tag')
    assert out[0].startswith('Usage: tag')
    assert out[1] == ''
    assert out[2] == 'create a html tag'


def test_argparse_prog(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'help tag')
    progname = out[0].split(' ')[1]
    assert progname == 'tag'


def test_arglist(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'arglist "we  should" get these')
    assert out[0] == 'True'


def test_arglist_kwargs(argparse_app, capsys) -> None:
    """Test with_argument_list wrapper passes through kwargs to command function"""
    argparse_app.do_arglist('arg', keyword_arg="foo")
    out, _err = capsys.readouterr()
    assert out == "foo\n"


def test_preservelist(argparse_app) -> None:
    out, _err = run_cmd(argparse_app, 'preservelist foo "bar baz"')
    assert out[0] == "['foo', '\"bar baz\"']"


def test_invalid_parser_builder(argparse_app):
    parser_builder = None
    with pytest.raises(TypeError):
        argparse_app._build_parser(argparse_app, parser_builder, "fake_prog")


def _build_has_subcmd_parser() -> cmd2.Cmd2ArgumentParser:
    has_subcmds_parser = cmd2.Cmd2ArgumentParser(description="Tests as_subcmd_to decorator")
    has_subcmds_parser.add_subparsers(dest='subcommand', metavar='SUBCOMMAND', required=True)
    return has_subcmds_parser


class SubcommandApp(cmd2.Cmd):
    """Example cmd2 application where we a base command which has a couple subcommands."""

    # subcommand functions for the base command
    def base_foo(self, args) -> None:
        """Foo subcommand of base command"""
        self.poutput(args.x * args.y)

    def base_bar(self, args) -> None:
        """Bar subcommand of base command"""
        self.poutput(f'(({args.z}))')

    def base_helpless(self, args) -> None:
        """Helpless subcommand of base command"""
        self.poutput(f'(({args.z}))')

    # create the top-level parser for the base command
    base_parser = cmd2.Cmd2ArgumentParser()
    base_subparsers = base_parser.add_subparsers(dest='subcommand', metavar='SUBCOMMAND', required=True)

    # create the parser for the "foo" subcommand
    parser_foo = base_subparsers.add_parser('foo', help='foo help')
    parser_foo.add_argument('-x', type=int, default=1, help='integer')
    parser_foo.add_argument('y', type=float, help='float')
    parser_foo.set_defaults(func=base_foo)

    # create the parser for the "bar" subcommand
    parser_bar = base_subparsers.add_parser('bar', help='bar help', aliases=['bar_1', 'bar_2'])
    parser_bar.add_argument('z', help='string')
    parser_bar.set_defaults(func=base_bar)

    # create the parser for the "helpless" subcommand
    # This subcommand has aliases and no help text. It exists to prevent changes to set_parser_prog() which
    # use an approach which relies on action._choices_actions list. See comment in that function for more
    # details.
    parser_helpless = base_subparsers.add_parser('helpless', aliases=['helpless_1', 'helpless_2'])
    parser_helpless.add_argument('z', help='string')
    parser_helpless.set_defaults(func=base_helpless)

    @cmd2.with_argparser(base_parser)
    def do_base(self, args) -> None:
        """Base command help"""
        # Call whatever subcommand function was selected
        func = args.func
        func(self, args)

    # Add subcommands using as_subcommand_to decorator
    @cmd2.with_argparser(_build_has_subcmd_parser)
    def do_test_subcmd_decorator(self, args: argparse.Namespace) -> None:
        handler = args.cmd2_handler.get()
        handler(args)

    subcmd_parser = cmd2.Cmd2ArgumentParser(description="A subcommand")

    @cmd2.as_subcommand_to('test_subcmd_decorator', 'subcmd', subcmd_parser, help=subcmd_parser.description.lower())
    def subcmd_func(self, args: argparse.Namespace) -> None:
        # Make sure printing the Namespace works. The way we originally added cmd2_handler to it resulted in a RecursionError.
        self.poutput(args)

    helpless_subcmd_parser = cmd2.Cmd2ArgumentParser(add_help=False, description="A subcommand with no help")

    @cmd2.as_subcommand_to(
        'test_subcmd_decorator', 'helpless_subcmd', helpless_subcmd_parser, help=helpless_subcmd_parser.description.lower()
    )
    def helpless_subcmd_func(self, args: argparse.Namespace) -> None:
        # Make sure vars(Namespace) works. The way we originally added cmd2_handler to it resulted in a RecursionError.
        self.poutput(vars(args))


@pytest.fixture
def subcommand_app():
    return SubcommandApp()


def test_subcommand_foo(subcommand_app) -> None:
    out, _err = run_cmd(subcommand_app, 'base foo -x2 5.0')
    assert out == ['10.0']


def test_subcommand_bar(subcommand_app) -> None:
    out, _err = run_cmd(subcommand_app, 'base bar baz')
    assert out == ['((baz))']


def test_subcommand_invalid(subcommand_app) -> None:
    _out, err = run_cmd(subcommand_app, 'base baz')
    assert err[0].startswith('Usage: base')
    assert err[1].startswith("Error: argument SUBCOMMAND: invalid choice: 'baz'")


def test_subcommand_base_help(subcommand_app) -> None:
    out, _err = run_cmd(subcommand_app, 'help base')
    assert out[0].startswith('Usage: base')
    assert out[1] == ''
    assert out[2] == 'Base command help'


def test_subcommand_help(subcommand_app) -> None:
    # foo has no aliases
    out, _err = run_cmd(subcommand_app, 'help base foo')
    assert out[0].startswith('Usage: base foo')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    # bar has aliases (usage should never show alias name)
    out, _err = run_cmd(subcommand_app, 'help base bar')
    assert out[0].startswith('Usage: base bar')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    out, _err = run_cmd(subcommand_app, 'help base bar_1')
    assert out[0].startswith('Usage: base bar')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    out, _err = run_cmd(subcommand_app, 'help base bar_2')
    assert out[0].startswith('Usage: base bar')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    # helpless has aliases and no help text (usage should never show alias name)
    out, _err = run_cmd(subcommand_app, 'help base helpless')
    assert out[0].startswith('Usage: base helpless')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    out, _err = run_cmd(subcommand_app, 'help base helpless_1')
    assert out[0].startswith('Usage: base helpless')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    out, _err = run_cmd(subcommand_app, 'help base helpless_2')
    assert out[0].startswith('Usage: base helpless')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'


def test_subcommand_invalid_help(subcommand_app) -> None:
    out, _err = run_cmd(subcommand_app, 'help base baz')
    assert out[0].startswith('Usage: base')


def test_add_another_subcommand(subcommand_app) -> None:
    """This tests makes sure set_parser_prog() sets _prog_prefix on every _SubParsersAction so that all future calls
    to add_parser() write the correct prog value to the parser being added.
    """
    base_parser = subcommand_app._command_parsers.get(subcommand_app.do_base)
    for sub_action in base_parser._actions:
        if isinstance(sub_action, argparse._SubParsersAction):
            new_parser = sub_action.add_parser('new_sub', help='stuff')
            break

    assert new_parser.prog == "base new_sub"


def test_subcmd_decorator(subcommand_app) -> None:
    # Test subcommand that has help option
    out, err = run_cmd(subcommand_app, 'test_subcmd_decorator subcmd')
    assert out[0].startswith('Namespace(')

    out, err = run_cmd(subcommand_app, 'help test_subcmd_decorator subcmd')
    assert out[0] == 'Usage: test_subcmd_decorator subcmd [-h]'

    out, err = run_cmd(subcommand_app, 'test_subcmd_decorator subcmd -h')
    assert out[0] == 'Usage: test_subcmd_decorator subcmd [-h]'

    # Test subcommand that has no help option
    out, err = run_cmd(subcommand_app, 'test_subcmd_decorator helpless_subcmd')
    assert "'subcommand': 'helpless_subcmd'" in out[1]

    out, err = run_cmd(subcommand_app, 'help test_subcmd_decorator helpless_subcmd')
    assert out[0] == 'Usage: test_subcmd_decorator helpless_subcmd'
    assert not err

    out, err = run_cmd(subcommand_app, 'test_subcmd_decorator helpless_subcmd -h')
    assert not out
    assert err[0] == 'Usage: test_subcmd_decorator [-h] SUBCOMMAND ...'
    assert err[1] == 'Error: unrecognized arguments: -h'


def test_unittest_mock() -> None:
    from unittest import (
        mock,
    )

    from cmd2 import (
        CommandSetRegistrationError,
    )

    with mock.patch.object(ArgparseApp, 'namespace_provider'), pytest.raises(CommandSetRegistrationError):
        ArgparseApp()

    with mock.patch.object(ArgparseApp, 'namespace_provider', spec=True):
        ArgparseApp()

    with mock.patch.object(ArgparseApp, 'namespace_provider', spec_set=True):
        ArgparseApp()

    with mock.patch.object(ArgparseApp, 'namespace_provider', autospec=True):
        ArgparseApp()


def test_pytest_mock_invalid(mocker) -> None:
    from cmd2 import (
        CommandSetRegistrationError,
    )

    mocker.patch.object(ArgparseApp, 'namespace_provider')
    with pytest.raises(CommandSetRegistrationError):
        ArgparseApp()


@pytest.mark.parametrize(
    'spec_param',
    [
        {'spec': True},
        {'spec_set': True},
        {'autospec': True},
    ],
)
def test_pytest_mock_valid(mocker, spec_param) -> None:
    mocker.patch.object(ArgparseApp, 'namespace_provider', **spec_param)
    ArgparseApp()



# Tests that cmd2 populates __name__, __file__, and sets sys.path[0] to our directory
import os
import sys

app.cmd_echo = True

if __name__ != '__main__':
    print(f"Error: __name__ is: {__name__}")
    quit()

if __file__ != sys.argv[0]:
    print(f"Error: __file__ is: {__file__}")
    quit()

our_dir = os.path.dirname(os.path.abspath(__file__))
if our_dir != sys.path[0]:
    print(f"Error: our_dir is: {our_dir}")
    quit()

print("PASSED")



from __future__ import annotations

import cmd2

from .conftest import normalize, run_cmd


def test_hooks_work_with_future_annotations() -> None:
    class HookApp(cmd2.Cmd):
        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args, **kwargs)
            self.register_cmdfinalization_hook(self.hook)

        def hook(self: cmd2.Cmd, data: cmd2.plugin.CommandFinalizationData) -> cmd2.plugin.CommandFinalizationData:
            if self.in_script():
                self.poutput("WE ARE IN SCRIPT")
            return data

    hook_app = HookApp()
    out, _err = run_cmd(hook_app, '')
    expected = normalize('')
    assert out == expected



#!/usr/bin/env python
"""A sample application for cmd2 demonstrating how to remove one of the built-in runtime settable parameters."""

import cmd2


class MyApp(cmd2.Cmd):
    def __init__(self) -> None:
        super().__init__()
        self.remove_settable('debug')


if __name__ == '__main__':
    import sys

    c = MyApp()
    sys.exit(c.cmdloop())



"""reproduces test_argparse.py except with SubCommands"""

import pytest

import cmd2

from .conftest import (
    WithCommandSets,
    run_cmd,
)


class SubcommandSet(cmd2.CommandSet):
    """Example cmd2 application where we a base command which has a couple subcommands."""

    def __init__(self, dummy) -> None:
        super().__init__()

    # subcommand functions for the base command
    def base_foo(self, args) -> None:
        """Foo subcommand of base command"""
        self._cmd.poutput(args.x * args.y)

    def base_bar(self, args) -> None:
        """Bar subcommand of base command"""
        self._cmd.poutput(f'(({args.z}))')

    def base_helpless(self, args) -> None:
        """Helpless subcommand of base command"""
        self._cmd.poutput(f'(({args.z}))')

    # create the top-level parser for the base command
    base_parser = cmd2.Cmd2ArgumentParser()
    base_subparsers = base_parser.add_subparsers(dest='subcommand', metavar='SUBCOMMAND', required=True)

    # create the parser for the "foo" subcommand
    parser_foo = base_subparsers.add_parser('foo', help='foo help')
    parser_foo.add_argument('-x', type=int, default=1, help='integer')
    parser_foo.add_argument('y', type=float, help='float')
    parser_foo.set_defaults(func=base_foo)

    # create the parser for the "bar" subcommand
    parser_bar = base_subparsers.add_parser('bar', help='bar help', aliases=['bar_1', 'bar_2'])
    parser_bar.add_argument('z', help='string')
    parser_bar.set_defaults(func=base_bar)

    # create the parser for the "helpless" subcommand
    # This subcommand has aliases and no help text. It exists to prevent changes to set_parser_prog() which
    # use an approach which relies on action._choices_actions list. See comment in that function for more
    # details.
    parser_helpless = base_subparsers.add_parser('helpless', aliases=['helpless_1', 'helpless_2'])
    parser_helpless.add_argument('z', help='string')
    parser_helpless.set_defaults(func=base_helpless)

    @cmd2.with_argparser(base_parser)
    def do_base(self, args) -> None:
        """Base command help"""
        # Call whatever subcommand function was selected
        func = args.func
        func(self, args)


@pytest.fixture
def subcommand_app():
    return WithCommandSets(auto_load_commands=False, command_sets=[SubcommandSet(1)])


def test_subcommand_foo(subcommand_app) -> None:
    out, _err = run_cmd(subcommand_app, 'base foo -x2 5.0')
    assert out == ['10.0']


def test_subcommand_bar(subcommand_app) -> None:
    out, _err = run_cmd(subcommand_app, 'base bar baz')
    assert out == ['((baz))']


def test_subcommand_invalid(subcommand_app) -> None:
    _out, err = run_cmd(subcommand_app, 'base baz')
    assert err[0].startswith('Usage: base')
    assert err[1].startswith("Error: argument SUBCOMMAND: invalid choice: 'baz'")


def test_subcommand_base_help(subcommand_app) -> None:
    out, _err = run_cmd(subcommand_app, 'help base')
    assert out[0].startswith('Usage: base')
    assert out[1] == ''
    assert out[2] == 'Base command help'


def test_subcommand_help(subcommand_app) -> None:
    # foo has no aliases
    out, _err = run_cmd(subcommand_app, 'help base foo')
    assert out[0].startswith('Usage: base foo')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    # bar has aliases (usage should never show alias name)
    out, _err = run_cmd(subcommand_app, 'help base bar')
    assert out[0].startswith('Usage: base bar')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    out, _err = run_cmd(subcommand_app, 'help base bar_1')
    assert out[0].startswith('Usage: base bar')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    out, _err = run_cmd(subcommand_app, 'help base bar_2')
    assert out[0].startswith('Usage: base bar')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    # helpless has aliases and no help text (usage should never show alias name)
    out, _err = run_cmd(subcommand_app, 'help base helpless')
    assert out[0].startswith('Usage: base helpless')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    out, _err = run_cmd(subcommand_app, 'help base helpless_1')
    assert out[0].startswith('Usage: base helpless')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'

    out, _err = run_cmd(subcommand_app, 'help base helpless_2')
    assert out[0].startswith('Usage: base helpless')
    assert out[1] == ''
    assert out[2] == 'Positional Arguments:'


def test_subcommand_invalid_help(subcommand_app) -> None:
    out, _err = run_cmd(subcommand_app, 'help base baz')
    assert out[0].startswith('Usage: base')



#!/usr/bin/env python3
"""A simple example demonstrating how do_* commands can be created in a loop."""

import functools

import cmd2
from cmd2.constants import (
    COMMAND_FUNC_PREFIX,
    HELP_FUNC_PREFIX,
)

COMMAND_LIST = ['foo', 'bar']
CATEGORY = 'Dynamic Commands'


class CommandsInLoop(cmd2.Cmd):
    """Example of dynamically adding do_* commands."""

    def __init__(self) -> None:
        # Add dynamic commands before calling cmd2.Cmd's init since it validates command names
        for command in COMMAND_LIST:
            # Create command function and add help category to it
            cmd_func = functools.partial(self.send_text, text=command)
            cmd2.categorize(cmd_func, CATEGORY)

            # Add command function to CLI object
            cmd_func_name = COMMAND_FUNC_PREFIX + command
            setattr(self, cmd_func_name, cmd_func)

            # Add help function to CLI object
            help_func = functools.partial(self.text_help, text=command)
            help_func_name = HELP_FUNC_PREFIX + command
            setattr(self, help_func_name, help_func)

        super().__init__(include_ipy=True)

    def send_text(self, _args: cmd2.Statement, *, text: str) -> None:
        """Simulate sending text to a server and printing the response."""
        self.poutput(text.capitalize())

    def text_help(self, *, text: str) -> None:
        """Deal with printing help for the dynamically added commands."""
        self.poutput(f"Simulate sending {text!r} to a server and printing the response")


if __name__ == '__main__':
    app = CommandsInLoop()
    app.cmdloop()



"""Test CommandSet."""

import argparse

import cmd2


@cmd2.with_default_category('Fruits')
class CommandSetA(cmd2.CommandSet):
    def do_apple(self, _statement: cmd2.Statement) -> None:
        self._cmd.poutput('Apple!')

    def do_banana(self, _statement: cmd2.Statement) -> None:
        """Banana Command."""
        self._cmd.poutput('Banana!!')

    cranberry_parser = cmd2.Cmd2ArgumentParser()
    cranberry_parser.add_argument('arg1', choices=['lemonade', 'juice', 'sauce'])

    @cmd2.with_argparser(cranberry_parser, with_unknown_args=True)
    def do_cranberry(self, ns: argparse.Namespace, unknown: list[str]) -> None:
        self._cmd.poutput(f'Cranberry {ns.arg1}!!')
        if unknown and len(unknown):
            self._cmd.poutput('Unknown: ' + ', '.join(['{}'] * len(unknown)).format(*unknown))
        self._cmd.last_result = {'arg1': ns.arg1, 'unknown': unknown}

    def help_cranberry(self) -> None:
        self._cmd.stdout.write('This command does diddly squat...\n')

    @cmd2.with_argument_list
    @cmd2.with_category('Also Alone')
    def do_durian(self, args: list[str]) -> None:
        """Durian Command."""
        self._cmd.poutput(f'{len(args)} Arguments: ')
        self._cmd.poutput(', '.join(['{}'] * len(args)).format(*args))

    def complete_durian(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        return self._cmd.basic_complete(text, line, begidx, endidx, ['stinks', 'smells', 'disgusting'])

    elderberry_parser = cmd2.Cmd2ArgumentParser()
    elderberry_parser.add_argument('arg1')

    @cmd2.with_category('Alone')
    @cmd2.with_argparser(elderberry_parser)
    def do_elderberry(self, ns: argparse.Namespace) -> None:
        self._cmd.poutput(f'Elderberry {ns.arg1}!!')



#!/usr/bin/env python
"""A sample application for cmd2 demonstrating customized environment parameters."""

import cmd2


class EnvironmentApp(cmd2.Cmd):
    """Example cmd2 application."""

    def __init__(self) -> None:
        super().__init__()
        self.degrees_c = 22
        self.sunny = False
        self.add_settable(
            cmd2.Settable('degrees_c', int, 'Temperature in Celsius', self, onchange_cb=self._onchange_degrees_c)
        )
        self.add_settable(cmd2.Settable('sunny', bool, 'Is it sunny outside?', self))

    def do_sunbathe(self, _arg) -> None:
        """Attempt to sunbathe."""
        if self.degrees_c < 20:
            result = f"It's {self.degrees_c} C - are you a penguin?"
        elif not self.sunny:
            result = 'Too dim.'
        else:
            result = 'UV is bad for your skin.'
        self.poutput(result)

    def _onchange_degrees_c(self, _param_name, _old, new) -> None:
        # if it's over 40C, it's gotta be sunny, right?
        if new > 40:
            self.sunny = True


if __name__ == '__main__':
    import sys

    c = EnvironmentApp()
    sys.exit(c.cmdloop())



"""Some useful argument types.

Note that these types can be used with other argparse-compatible libraries, including
"argparse" itself.

The 'type' parameter to ArgumentParser.add_argument() must be a callable object,
typically a function. That function is called to convert the string to the Python type available
in the 'namespace' passed to your "do_xyz" command function. Thus, "type=int" works because
int("53") returns the integer value 53. If that callable object / function raises an exception
due to invalid input, the name ("repr") of the object/function will be printed in the error message
to the user. Using lambda, functools.partial, or the like will generate a callable object with a
rather opaque repr so it can be useful to have a one-line function rather than relying on a lambda,
even for a short expression.

For "types" that have some context/state, using a class with a __call__ method, and overriding
the __repr__ method, allows you to produce an error message that provides that information
to the user.
"""

from collections.abc import Iterable

import cmd2

_int_suffixes = {
    # SI number suffixes (unit prefixes):
    "K": 1_000,
    "M": 1_000_000,
    "G": 1_000_000_000,
    "T": 1_000_000_000_000,
    "P": 1_000_000_000_000_000,
    # IEC number suffixes (unit prefixes):
    "Ki": 1024,
    "Mi": 1024 * 1024,
    "Gi": 1024 * 1024 * 1024,
    "Ti": 1024 * 1024 * 1024 * 1024,
    "Pi": 1024 * 1024 * 1024 * 1024 * 1024,
}


def integer(value_str: str) -> int:
    """Will accept any base, and optional suffix like '64K'."""
    multiplier = 1
    # If there is a matching suffix, use its multiplier:
    for suffix, suffix_multiplier in _int_suffixes.items():
        if value_str.endswith(suffix):
            value_str = value_str.removesuffix(suffix)
            multiplier = suffix_multiplier
            break

    return int(value_str, 0) * multiplier


def hexadecimal(value_str: str) -> int:
    """Parse hexidecimal integer, with optional '0x' prefix."""
    return int(value_str, base=16)


class Range:
    """Useful as type for large ranges, when 'choices=range(maxval)' would be excessively large."""

    def __init__(self, firstval: int, secondval: int | None = None) -> None:
        """Construct a Range, with same syntax as 'range'.

        :param firstval: either the top end of range (if 'secondval' is missing), or the bottom end
        :param secondval: top end of range (one higher than maximum value)
        """
        if secondval is None:
            self.bottom = 0
            self.top = firstval
        else:
            self.bottom = firstval
            self.top = secondval

        self.range_str = f"[{self.bottom}..{self.top - 1}]"

    def __repr__(self) -> str:
        """Will be printed as the 'argument type' to user on syntax or range error."""
        return f"Range{self.range_str}"

    def __call__(self, arg: str) -> int:
        """Parse the string argument and checks validity."""
        val = integer(arg)
        if self.bottom <= val < self.top:
            return val
        raise ValueError(f"Value '{val}' not within {self.range_str}")


class IntSet:
    """Set of integers from a specified range.

    e.g. '5', '1-3,8', 'all'
    """

    def __init__(self, firstval: int, secondval: int | None = None) -> None:
        """Construct an IntSet, with same syntax as 'range'.

        :param firstval: either the top end of range (if 'secondval' is missing), or the bottom end
        :param secondval: top end of range (one higher than maximum value)
        """
        if secondval is None:
            self.bottom = 0
            self.top = firstval
        else:
            self.bottom = firstval
            self.top = secondval

        self.range_str = f"[{self.bottom}..{self.top - 1}]"

    def __repr__(self) -> str:
        """Will be printed as the 'argument type' to user on syntax or range error."""
        return f"IntSet{self.range_str}"

    def __call__(self, arg: str) -> Iterable[int]:
        """Parse a string into an iterable returning ints."""
        if arg == 'all':
            return range(self.bottom, self.top)

        out = []
        for piece in arg.split(','):
            if '-' in piece:
                a, b = [int(x) for x in piece.split('-', 2)]
                if a < self.bottom:
                    raise ValueError(f"Value '{a}' not within {self.range_str}")
                if b >= self.top:
                    raise ValueError(f"Value '{b}' not within {self.range_str}")
                out += list(range(a, b + 1))
            else:
                val = int(piece)
                if not self.bottom <= val < self.top:
                    raise ValueError(f"Value '{val}' not within {self.range_str}")
                out += [val]
        return out


if __name__ == '__main__':
    import argparse
    import sys

    class CustomTypesExample(cmd2.Cmd):
        example_parser = cmd2.Cmd2ArgumentParser()
        example_parser.add_argument(
            '--value', '-v', type=integer, help='Integer value, with optional K/M/G/Ki/Mi/Gi/... suffix'
        )
        example_parser.add_argument('--memory-address', '-m', type=hexadecimal, help='Memory address in hex')
        example_parser.add_argument('--year', type=Range(1900, 2000), help='Year between 1900-1999')
        example_parser.add_argument(
            '--index', dest='index_list', type=IntSet(100), help='One or more indexes 0-99. e.g. "1,3,5", "10,30-50", "all"'
        )

        @cmd2.with_argparser(example_parser)
        def do_example(self, args: argparse.Namespace) -> None:
            """The example command."""
            if args.value is not None:
                self.poutput(f"Value: {args.value}")
            if args.memory_address is not None:
                # print the value as hex, with leading "0x" + 16 hex digits + three '_' group separators:
                self.poutput(f"Address: {args.memory_address:#021_x}")
            if args.year is not None:
                self.poutput(f"Year: {args.year}")
            if args.index_list is not None:
                for index in args.index_list:
                    self.poutput(f"Process index {index}")

    app = CustomTypesExample()
    sys.exit(app.cmdloop())



#!/usr/bin/env python3
"""A simple example of setting a custom theme for a cmd2 application."""

from rich.style import Style

import cmd2
import cmd2.rich_utils as ru
from cmd2 import Cmd2Style, Color


class ThemedApp(cmd2.Cmd):
    """A simple cmd2 application with a custom theme."""

    def __init__(self, *args, **kwargs):
        """Initialize the application."""
        super().__init__(*args, **kwargs)
        self.intro = "This is a themed application. Try the 'theme_show' command."

        # Set text which prints right before all of the help tables are listed.
        self.doc_leader = "Welcome to this glorious help ..."

        # Create a custom theme
        # Colors can come from the cmd2.color.Color StrEnum class, be RGB hex values, or
        # be any of the rich standard colors: https://rich.readthedocs.io/en/stable/appendix/colors.html
        custom_theme = {
            Cmd2Style.SUCCESS: Style(color=Color.GREEN1, bgcolor=Color.GRAY30),  # Use color from cmd2 Color class
            Cmd2Style.WARNING: Style(color=Color.ORANGE1),
            Cmd2Style.ERROR: Style(color=Color.PINK1),
            Cmd2Style.HELP_HEADER: Style(color=Color.CYAN, bgcolor="#44475a"),
            Cmd2Style.HELP_LEADER: Style(color="#f8f8f2", bgcolor="#282a36"),  # use RGB hex colors
            Cmd2Style.TABLE_BORDER: Style(color="turquoise2"),  # use a rich standard color
            "traceback.exc_type": Style(color=Color.RED, bgcolor=Color.LIGHT_YELLOW3, bold=True),
            "argparse.args": Style(color=Color.AQUAMARINE3, underline=True),
        }
        ru.set_theme(custom_theme)

    @cmd2.with_category("Theme Commands")
    def do_theme_show(self, _: cmd2.Statement):
        """Showcases the custom theme by printing messages with different styles."""
        self.poutput("This is a basic output message.")
        self.psuccess("This is a success message.")
        self.pwarning("This is a warning message.")
        self.perror("This is an error message.")
        self.pexcept(ValueError("This is a dummy ValueError exception."))


if __name__ == "__main__":
    import sys

    app = ThemedApp()
    sys.exit(app.cmdloop())



#!/usr/bin/env python
"""A sample application for tagging categories on commands.

It also demonstrates the effects of decorator order when it comes to argparse errors occurring.
"""

import functools

import cmd2
from cmd2 import COMMAND_NAME


def my_decorator(f):
    @functools.wraps(f)
    def wrapper(*args, **kwds):
        print('Calling decorated function')
        return f(*args, **kwds)

    return wrapper


class HelpCategories(cmd2.Cmd):
    """Example cmd2 application."""

    START_TIMES = ('now', 'later', 'sometime', 'whenever')

    # Command categories
    CMD_CAT_CONNECTING = 'Connecting'
    CMD_CAT_APP_MGMT = 'Application Management'
    CMD_CAT_SERVER_INFO = 'Server Information'

    def __init__(self) -> None:
        super().__init__()

        # Set the default category for uncategorized commands
        self.default_category = 'Other'

    def do_connect(self, _) -> None:
        """Connect command."""
        self.poutput('Connect')

    # Tag the above command functions under the category Connecting
    cmd2.categorize(do_connect, CMD_CAT_CONNECTING)

    @cmd2.with_category(CMD_CAT_CONNECTING)
    def do_which(self, _) -> None:
        """Which command."""
        self.poutput('Which')

    def do_list(self, _) -> None:
        """List command."""
        self.poutput('List')

    def do_deploy(self, _) -> None:
        """Deploy command."""
        self.poutput('Deploy')

    start_parser = cmd2.Cmd2ArgumentParser(
        description='Start',
        epilog='my_decorator runs even with argparse errors',
    )
    start_parser.add_argument('when', choices=START_TIMES, help='Specify when to start')

    @my_decorator
    @cmd2.with_argparser(start_parser)
    def do_start(self, _) -> None:
        """Start command."""
        self.poutput('Start')

    def do_sessions(self, _) -> None:
        """Sessions command."""
        self.poutput('Sessions')

    def do_redeploy(self, _) -> None:
        """Redeploy command."""
        self.poutput('Redeploy')

    restart_parser = cmd2.Cmd2ArgumentParser(
        description='Restart',
        epilog='my_decorator does not run when argparse errors',
    )
    restart_parser.add_argument('when', choices=START_TIMES, help='Specify when to restart')

    @cmd2.with_argparser(restart_parser)
    @cmd2.with_category(CMD_CAT_APP_MGMT)
    @my_decorator
    def do_restart(self, _) -> None:
        """Restart command."""
        self.poutput('Restart')

    def do_expire(self, _) -> None:
        """Expire command."""
        self.poutput('Expire')

    def do_undeploy(self, _) -> None:
        """Undeploy command."""
        self.poutput('Undeploy')

    def do_stop(self, _) -> None:
        """Stop command."""
        self.poutput('Stop')

    def do_findleakers(self, _) -> None:
        """Find Leakers command."""
        self.poutput('Find Leakers')

    # Tag the above command functions under the category Application Management
    cmd2.categorize(
        (do_list, do_deploy, do_start, do_sessions, do_redeploy, do_expire, do_undeploy, do_stop, do_findleakers),
        CMD_CAT_APP_MGMT,
    )

    def do_resources(self, _) -> None:
        """Resources command."""
        self.poutput('Resources')

    def do_status(self, _) -> None:
        """Status command."""
        self.poutput('Status')

    def do_serverinfo(self, _) -> None:
        """Server Info command."""
        self.poutput('Server Info')

    def do_thread_dump(self, _) -> None:
        """Thread Dump command."""
        self.poutput('Thread Dump')

    def do_sslconnectorciphers(self, _) -> None:
        """SSL Connector Ciphers command is an example of a command that contains
        multiple lines of help information for the user. Each line of help in a
        contiguous set of lines will be printed and aligned in the verbose output
        provided with 'help --verbose'.

        This is after a blank line and won't de displayed in the verbose help
        """
        self.poutput('SSL Connector Ciphers')

    def do_vminfo(self, _) -> None:
        """VM Info command."""
        self.poutput('VM Info')

    # Tag the above command functions under the category Server Information
    cmd2.categorize(do_resources, CMD_CAT_SERVER_INFO)
    cmd2.categorize(do_status, CMD_CAT_SERVER_INFO)
    cmd2.categorize(do_serverinfo, CMD_CAT_SERVER_INFO)
    cmd2.categorize(do_thread_dump, CMD_CAT_SERVER_INFO)
    cmd2.categorize(do_sslconnectorciphers, CMD_CAT_SERVER_INFO)
    cmd2.categorize(do_vminfo, CMD_CAT_SERVER_INFO)

    # The following command functions don't have the HELP_CATEGORY attribute set
    # and show up in the 'Other' group
    def do_config(self, _) -> None:
        """Config command."""
        self.poutput('Config')

    def do_version(self, _) -> None:
        """Version command."""
        self.poutput(cmd2.__version__)

    @cmd2.with_category("Command Management")
    def do_disable_commands(self, _) -> None:
        """Disable the Application Management commands."""
        message_to_print = f"{COMMAND_NAME} is not available while {self.CMD_CAT_APP_MGMT} commands are disabled"
        self.disable_category(self.CMD_CAT_APP_MGMT, message_to_print)
        self.poutput("The Application Management commands have been disabled")

    @cmd2.with_category("Command Management")
    def do_enable_commands(self, _) -> None:
        """Enable the Application Management commands."""
        self.enable_category(self.CMD_CAT_APP_MGMT)
        self.poutput("The Application Management commands have been enabled")


if __name__ == '__main__':
    import sys

    c = HelpCategories()
    sys.exit(c.cmdloop())



"""Test the parsing logic in parsing.py"""

import dataclasses

import pytest

import cmd2
from cmd2 import (
    constants,
    exceptions,
)
from cmd2 import string_utils as su
from cmd2.parsing import (
    Statement,
    StatementParser,
    shlex_split,
)


@pytest.fixture
def parser():
    return StatementParser(
        terminators=[';', '&'],
        multiline_commands=['multiline'],
        aliases={
            'helpalias': 'help',
            '42': 'theanswer',
            'l': '!ls -al',
            'anothermultiline': 'multiline',
            'fake': 'run_pyscript',
        },
        shortcuts={'?': 'help', '!': 'shell'},
    )


@pytest.fixture
def default_parser():
    return StatementParser()


def test_parse_empty_string(parser) -> None:
    line = ''
    statement = parser.parse(line)
    assert statement == ''
    assert statement.args == statement
    assert statement.raw == line
    assert statement.command == ''
    assert statement.arg_list == []
    assert statement.multiline_command == ''
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == ''
    assert statement.output_to == ''
    assert statement.command_and_args == line
    assert statement.argv == statement.arg_list


def test_parse_empty_string_default(default_parser) -> None:
    line = ''
    statement = default_parser.parse(line)
    assert statement == ''
    assert statement.args == statement
    assert statement.raw == line
    assert statement.command == ''
    assert statement.arg_list == []
    assert statement.multiline_command == ''
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == ''
    assert statement.output_to == ''
    assert statement.command_and_args == line
    assert statement.argv == statement.arg_list


@pytest.mark.parametrize(
    ('line', 'tokens'),
    [
        ('command', ['command']),
        (constants.COMMENT_CHAR + 'comment', []),
        ('not ' + constants.COMMENT_CHAR + ' a comment', ['not', constants.COMMENT_CHAR, 'a', 'comment']),
        ('termbare ; > /tmp/output', ['termbare', ';', '>', '/tmp/output']),
        ('termbare; > /tmp/output', ['termbare', ';', '>', '/tmp/output']),
        ('termbare & > /tmp/output', ['termbare', '&', '>', '/tmp/output']),
        ('termbare& > /tmp/output', ['termbare&', '>', '/tmp/output']),
        ('help|less', ['help', '|', 'less']),
    ],
)
def test_tokenize_default(default_parser, line, tokens) -> None:
    tokens_to_test = default_parser.tokenize(line)
    assert tokens_to_test == tokens


@pytest.mark.parametrize(
    ('line', 'tokens'),
    [
        ('command', ['command']),
        ('# comment', []),
        ('not ' + constants.COMMENT_CHAR + ' a comment', ['not', constants.COMMENT_CHAR, 'a', 'comment']),
        ('42 arg1 arg2', ['theanswer', 'arg1', 'arg2']),
        ('l', ['shell', 'ls', '-al']),
        ('termbare ; > /tmp/output', ['termbare', ';', '>', '/tmp/output']),
        ('termbare; > /tmp/output', ['termbare', ';', '>', '/tmp/output']),
        ('termbare & > /tmp/output', ['termbare', '&', '>', '/tmp/output']),
        ('termbare& > /tmp/output', ['termbare', '&', '>', '/tmp/output']),
        ('help|less', ['help', '|', 'less']),
        ('l|less', ['shell', 'ls', '-al', '|', 'less']),
    ],
)
def test_tokenize(parser, line, tokens) -> None:
    tokens_to_test = parser.tokenize(line)
    assert tokens_to_test == tokens


def test_tokenize_unclosed_quotes(parser) -> None:
    with pytest.raises(exceptions.Cmd2ShlexError):
        _ = parser.tokenize('command with "unclosed quotes')


@pytest.mark.parametrize(
    ('tokens', 'command', 'args'),
    [([], '', ''), (['command'], 'command', ''), (['command', 'arg1', 'arg2'], 'command', 'arg1 arg2')],
)
def test_command_and_args(parser, tokens, command, args) -> None:
    (parsed_command, parsed_args) = parser._command_and_args(tokens)
    assert command == parsed_command
    assert args == parsed_args


@pytest.mark.parametrize(
    'line',
    [
        'plainword',
        '"one word"',
        "'one word'",
    ],
)
def test_parse_single_word(parser, line) -> None:
    statement = parser.parse(line)
    assert statement.command == line
    assert statement == ''
    assert statement.argv == [su.strip_quotes(line)]
    assert not statement.arg_list
    assert statement.args == statement
    assert statement.raw == line
    assert statement.multiline_command == ''
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == ''
    assert statement.output_to == ''
    assert statement.command_and_args == line


@pytest.mark.parametrize(
    ('line', 'terminator'),
    [
        ('termbare;', ';'),
        ('termbare ;', ';'),
        ('termbare&', '&'),
        ('termbare &', '&'),
    ],
)
def test_parse_word_plus_terminator(parser, line, terminator) -> None:
    statement = parser.parse(line)
    assert statement.command == 'termbare'
    assert statement == ''
    assert statement.argv == ['termbare']
    assert not statement.arg_list
    assert statement.terminator == terminator
    assert statement.expanded_command_line == statement.command + statement.terminator


@pytest.mark.parametrize(
    ('line', 'terminator'),
    [
        ('termbare;  suffx', ';'),
        ('termbare ;suffx', ';'),
        ('termbare&  suffx', '&'),
        ('termbare &suffx', '&'),
    ],
)
def test_parse_suffix_after_terminator(parser, line, terminator) -> None:
    statement = parser.parse(line)
    assert statement.command == 'termbare'
    assert statement == ''
    assert statement.args == statement
    assert statement.argv == ['termbare']
    assert not statement.arg_list
    assert statement.terminator == terminator
    assert statement.suffix == 'suffx'
    assert statement.expanded_command_line == statement.command + statement.terminator + ' ' + statement.suffix


def test_parse_command_with_args(parser) -> None:
    line = 'command with args'
    statement = parser.parse(line)
    assert statement.command == 'command'
    assert statement == 'with args'
    assert statement.args == statement
    assert statement.argv == ['command', 'with', 'args']
    assert statement.arg_list == statement.argv[1:]


def test_parse_command_with_quoted_args(parser) -> None:
    line = 'command with "quoted args" and "some not"'
    statement = parser.parse(line)
    assert statement.command == 'command'
    assert statement == 'with "quoted args" and "some not"'
    assert statement.args == statement
    assert statement.argv == ['command', 'with', 'quoted args', 'and', 'some not']
    assert statement.arg_list == ['with', '"quoted args"', 'and', '"some not"']


def test_parse_command_with_args_terminator_and_suffix(parser) -> None:
    line = 'command with args and terminator; and suffix'
    statement = parser.parse(line)
    assert statement.command == 'command'
    assert statement == "with args and terminator"
    assert statement.args == statement
    assert statement.argv == ['command', 'with', 'args', 'and', 'terminator']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ';'
    assert statement.suffix == 'and suffix'


def test_parse_comment(parser) -> None:
    statement = parser.parse(constants.COMMENT_CHAR + ' this is all a comment')
    assert statement.command == ''
    assert statement == ''
    assert statement.args == statement
    assert not statement.argv
    assert not statement.arg_list


def test_parse_embedded_comment_char(parser) -> None:
    command_str = 'hi ' + constants.COMMENT_CHAR + ' not a comment'
    statement = parser.parse(command_str)
    assert statement.command == 'hi'
    assert statement == constants.COMMENT_CHAR + ' not a comment'
    assert statement.args == statement
    assert statement.argv == shlex_split(command_str)
    assert statement.arg_list == statement.argv[1:]


@pytest.mark.parametrize(
    'line',
    [
        'simple | piped',
        'simple|piped',
    ],
)
def test_parse_simple_pipe(parser, line) -> None:
    statement = parser.parse(line)
    assert statement.command == 'simple'
    assert statement == ''
    assert statement.args == statement
    assert statement.argv == ['simple']
    assert not statement.arg_list
    assert statement.pipe_to == 'piped'
    assert statement.expanded_command_line == statement.command + ' | ' + statement.pipe_to


def test_parse_double_pipe_is_not_a_pipe(parser) -> None:
    line = 'double-pipe || is not a pipe'
    statement = parser.parse(line)
    assert statement.command == 'double-pipe'
    assert statement == '|| is not a pipe'
    assert statement.args == statement
    assert statement.argv == ['double-pipe', '||', 'is', 'not', 'a', 'pipe']
    assert statement.arg_list == statement.argv[1:]
    assert not statement.pipe_to


def test_parse_complex_pipe(parser) -> None:
    line = 'command with args, terminator&sufx | piped'
    statement = parser.parse(line)
    assert statement.command == 'command'
    assert statement == "with args, terminator"
    assert statement.args == statement
    assert statement.argv == ['command', 'with', 'args,', 'terminator']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == '&'
    assert statement.suffix == 'sufx'
    assert statement.pipe_to == 'piped'


@pytest.mark.parametrize(
    ('line', 'output'),
    [
        ('help > out.txt', '>'),
        ('help>out.txt', '>'),
        ('help >> out.txt', '>>'),
        ('help>>out.txt', '>>'),
    ],
)
def test_parse_redirect(parser, line, output) -> None:
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == ''
    assert statement.args == statement
    assert statement.output == output
    assert statement.output_to == 'out.txt'
    assert statement.expanded_command_line == statement.command + ' ' + statement.output + ' ' + statement.output_to


@pytest.mark.parametrize(
    'dest',
    [
        'afile.txt',
        'python-cmd2/afile.txt',
    ],
)  # without dashes  # with dashes in path
def test_parse_redirect_with_args(parser, dest) -> None:
    line = f'output into > {dest}'
    statement = parser.parse(line)
    assert statement.command == 'output'
    assert statement == 'into'
    assert statement.args == statement
    assert statement.argv == ['output', 'into']
    assert statement.arg_list == statement.argv[1:]
    assert statement.output == '>'
    assert statement.output_to == dest


def test_parse_redirect_append(parser) -> None:
    line = 'output appended to >> /tmp/afile.txt'
    statement = parser.parse(line)
    assert statement.command == 'output'
    assert statement == 'appended to'
    assert statement.args == statement
    assert statement.argv == ['output', 'appended', 'to']
    assert statement.arg_list == statement.argv[1:]
    assert statement.output == '>>'
    assert statement.output_to == '/tmp/afile.txt'


def test_parse_pipe_then_redirect(parser) -> None:
    line = 'output into;sufx | pipethrume plz > afile.txt'
    statement = parser.parse(line)
    assert statement.command == 'output'
    assert statement == 'into'
    assert statement.args == statement
    assert statement.argv == ['output', 'into']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ';'
    assert statement.suffix == 'sufx'
    assert statement.pipe_to == 'pipethrume plz > afile.txt'
    assert statement.output == ''
    assert statement.output_to == ''


def test_parse_multiple_pipes(parser) -> None:
    line = 'output into;sufx | pipethrume plz | grep blah'
    statement = parser.parse(line)
    assert statement.command == 'output'
    assert statement == 'into'
    assert statement.args == statement
    assert statement.argv == ['output', 'into']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ';'
    assert statement.suffix == 'sufx'
    assert statement.pipe_to == 'pipethrume plz | grep blah'
    assert statement.output == ''
    assert statement.output_to == ''


def test_redirect_then_pipe(parser) -> None:
    line = 'help alias > file.txt | grep blah'
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == 'alias'
    assert statement.args == statement
    assert statement.argv == ['help', 'alias']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == '>'
    assert statement.output_to == 'file.txt'


def test_append_then_pipe(parser) -> None:
    line = 'help alias >> file.txt | grep blah'
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == 'alias'
    assert statement.args == statement
    assert statement.argv == ['help', 'alias']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == '>>'
    assert statement.output_to == 'file.txt'


def test_append_then_redirect(parser) -> None:
    line = 'help alias >> file.txt > file2.txt'
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == 'alias'
    assert statement.args == statement
    assert statement.argv == ['help', 'alias']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == '>>'
    assert statement.output_to == 'file.txt'


def test_redirect_then_append(parser) -> None:
    line = 'help alias > file.txt >> file2.txt'
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == 'alias'
    assert statement.args == statement
    assert statement.argv == ['help', 'alias']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == '>'
    assert statement.output_to == 'file.txt'


def test_redirect_to_quoted_string(parser) -> None:
    line = 'help alias > "file.txt"'
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == 'alias'
    assert statement.args == statement
    assert statement.argv == ['help', 'alias']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == '>'
    assert statement.output_to == '"file.txt"'


def test_redirect_to_single_quoted_string(parser) -> None:
    line = "help alias > 'file.txt'"
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == 'alias'
    assert statement.args == statement
    assert statement.argv == ['help', 'alias']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == '>'
    assert statement.output_to == "'file.txt'"


def test_redirect_to_empty_quoted_string(parser) -> None:
    line = 'help alias > ""'
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == 'alias'
    assert statement.args == statement
    assert statement.argv == ['help', 'alias']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == '>'
    assert statement.output_to == ''


def test_redirect_to_empty_single_quoted_string(parser) -> None:
    line = "help alias > ''"
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == 'alias'
    assert statement.args == statement
    assert statement.argv == ['help', 'alias']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == '>'
    assert statement.output_to == ''


def test_parse_output_to_paste_buffer(parser) -> None:
    line = 'output to paste buffer >> '
    statement = parser.parse(line)
    assert statement.command == 'output'
    assert statement == 'to paste buffer'
    assert statement.args == statement
    assert statement.argv == ['output', 'to', 'paste', 'buffer']
    assert statement.arg_list == statement.argv[1:]
    assert statement.output == '>>'


def test_parse_redirect_inside_terminator(parser) -> None:
    """The terminator designates the end of the command/arguments portion.
    If a redirector occurs before a terminator, then it will be treated as
    part of the arguments and not as a redirector.
    """
    line = 'has > inside;'
    statement = parser.parse(line)
    assert statement.command == 'has'
    assert statement == '> inside'
    assert statement.args == statement
    assert statement.argv == ['has', '>', 'inside']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ';'


@pytest.mark.parametrize(
    ('line', 'terminator'),
    [
        ('multiline with | inside;', ';'),
        ('multiline with | inside ;', ';'),
        ('multiline with | inside;;;', ';'),
        ('multiline with | inside;; ;;', ';'),
        ('multiline with | inside&', '&'),
        ('multiline with | inside &;', '&'),
        ('multiline with | inside&&;', '&'),
        ('multiline with | inside &; &;', '&'),
    ],
)
def test_parse_multiple_terminators(parser, line, terminator) -> None:
    statement = parser.parse(line)
    assert statement.multiline_command == 'multiline'
    assert statement == 'with | inside'
    assert statement.args == statement
    assert statement.argv == ['multiline', 'with', '|', 'inside']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == terminator


def test_parse_unfinished_multiliine_command(parser) -> None:
    line = 'multiline has > inside an unfinished command'
    statement = parser.parse(line)
    assert statement.multiline_command == 'multiline'
    assert statement.command == 'multiline'
    assert statement == 'has > inside an unfinished command'
    assert statement.args == statement
    assert statement.argv == ['multiline', 'has', '>', 'inside', 'an', 'unfinished', 'command']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == ''


def test_parse_basic_multiline_command(parser) -> None:
    line = 'multiline foo\nbar\n\n'
    statement = parser.parse(line)
    assert statement.multiline_command == 'multiline'
    assert statement.command == 'multiline'
    assert statement == 'foo bar'
    assert statement.args == statement
    assert statement.argv == ['multiline', 'foo', 'bar']
    assert statement.arg_list == ['foo', 'bar']
    assert statement.raw == line
    assert statement.terminator == '\n'


@pytest.mark.parametrize(
    ('line', 'terminator'),
    [
        ('multiline has > inside;', ';'),
        ('multiline has > inside;;;', ';'),
        ('multiline has > inside;; ;;', ';'),
        ('multiline has > inside &', '&'),
        ('multiline has > inside & &', '&'),
    ],
)
def test_parse_multiline_command_ignores_redirectors_within_it(parser, line, terminator) -> None:
    statement = parser.parse(line)
    assert statement.multiline_command == 'multiline'
    assert statement == 'has > inside'
    assert statement.args == statement
    assert statement.argv == ['multiline', 'has', '>', 'inside']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == terminator


def test_parse_multiline_terminated_by_empty_line(parser) -> None:
    line = 'multiline command ends\n\n'
    statement = parser.parse(line)
    assert statement.multiline_command == 'multiline'
    assert statement.command == 'multiline'
    assert statement == 'command ends'
    assert statement.args == statement
    assert statement.argv == ['multiline', 'command', 'ends']
    assert statement.arg_list == statement.argv[1:]
    assert statement.terminator == '\n'


@pytest.mark.parametrize(
    ('line', 'terminator'),
    [
        ('multiline command "with\nembedded newline";', ';'),
        ('multiline command "with\nembedded newline";;;', ';'),
        ('multiline command "with\nembedded newline";; ;;', ';'),
        ('multiline command "with\nembedded newline" &', '&'),
        ('multiline command "with\nembedded newline" & &', '&'),
        ('multiline command "with\nembedded newline"\n\n', '\n'),
    ],
)
def test_parse_multiline_with_embedded_newline(parser, line, terminator) -> None:
    statement = parser.parse(line)
    assert statement.multiline_command == 'multiline'
    assert statement.command == 'multiline'
    assert statement == 'command "with\nembedded newline"'
    assert statement.args == statement
    assert statement.argv == ['multiline', 'command', 'with\nembedded newline']
    assert statement.arg_list == ['command', '"with\nembedded newline"']
    assert statement.terminator == terminator


def test_parse_multiline_ignores_terminators_in_quotes(parser) -> None:
    line = 'multiline command "with term; ends" now\n\n'
    statement = parser.parse(line)
    assert statement.multiline_command == 'multiline'
    assert statement.command == 'multiline'
    assert statement == 'command "with term; ends" now'
    assert statement.args == statement
    assert statement.argv == ['multiline', 'command', 'with term; ends', 'now']
    assert statement.arg_list == ['command', '"with term; ends"', 'now']
    assert statement.terminator == '\n'


def test_parse_command_with_unicode_args(parser) -> None:
    line = 'drink café'
    statement = parser.parse(line)
    assert statement.command == 'drink'
    assert statement == 'café'
    assert statement.args == statement
    assert statement.argv == ['drink', 'café']
    assert statement.arg_list == statement.argv[1:]


def test_parse_unicode_command(parser) -> None:
    line = 'café au lait'
    statement = parser.parse(line)
    assert statement.command == 'café'
    assert statement == 'au lait'
    assert statement.args == statement
    assert statement.argv == ['café', 'au', 'lait']
    assert statement.arg_list == statement.argv[1:]


def test_parse_redirect_to_unicode_filename(parser) -> None:
    line = 'dir home > café'
    statement = parser.parse(line)
    assert statement.command == 'dir'
    assert statement == 'home'
    assert statement.args == statement
    assert statement.argv == ['dir', 'home']
    assert statement.arg_list == statement.argv[1:]
    assert statement.output == '>'
    assert statement.output_to == 'café'


def test_parse_unclosed_quotes(parser) -> None:
    with pytest.raises(exceptions.Cmd2ShlexError):
        _ = parser.tokenize("command with 'unclosed quotes")


def test_empty_statement_raises_exception() -> None:
    app = cmd2.Cmd()
    with pytest.raises(exceptions.EmptyStatement):
        app._complete_statement('')

    with pytest.raises(exceptions.EmptyStatement):
        app._complete_statement(' ')


@pytest.mark.parametrize(
    ('line', 'command', 'args'),
    [
        ('helpalias', 'help', ''),
        ('helpalias mycommand', 'help', 'mycommand'),
        ('42', 'theanswer', ''),
        ('42 arg1 arg2', 'theanswer', 'arg1 arg2'),
        ('!ls', 'shell', 'ls'),
        ('!ls -al /tmp', 'shell', 'ls -al /tmp'),
        ('l', 'shell', 'ls -al'),
    ],
)
def test_parse_alias_and_shortcut_expansion(parser, line, command, args) -> None:
    statement = parser.parse(line)
    assert statement.command == command
    assert statement == args
    assert statement.args == statement


def test_parse_alias_on_multiline_command(parser) -> None:
    line = 'anothermultiline has > inside an unfinished command'
    statement = parser.parse(line)
    assert statement.multiline_command == 'multiline'
    assert statement.command == 'multiline'
    assert statement.args == statement
    assert statement == 'has > inside an unfinished command'
    assert statement.terminator == ''


@pytest.mark.parametrize(
    ('line', 'output'),
    [
        ('helpalias > out.txt', '>'),
        ('helpalias>out.txt', '>'),
        ('helpalias >> out.txt', '>>'),
        ('helpalias>>out.txt', '>>'),
    ],
)
def test_parse_alias_redirection(parser, line, output) -> None:
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == ''
    assert statement.args == statement
    assert statement.output == output
    assert statement.output_to == 'out.txt'


@pytest.mark.parametrize(
    'line',
    [
        'helpalias | less',
        'helpalias|less',
    ],
)
def test_parse_alias_pipe(parser, line) -> None:
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == ''
    assert statement.args == statement
    assert statement.pipe_to == 'less'


@pytest.mark.parametrize(
    'line',
    [
        'helpalias;',
        'helpalias;;',
        'helpalias;; ;',
        'helpalias ;',
        'helpalias ; ;',
        'helpalias ;; ;',
    ],
)
def test_parse_alias_terminator_no_whitespace(parser, line) -> None:
    statement = parser.parse(line)
    assert statement.command == 'help'
    assert statement == ''
    assert statement.args == statement
    assert statement.terminator == ';'


def test_parse_command_only_command_and_args(parser) -> None:
    line = 'help history'
    statement = parser.parse_command_only(line)
    assert statement == 'history'
    assert statement.args == statement
    assert statement.arg_list == []
    assert statement.command == 'help'
    assert statement.command_and_args == line
    assert statement.multiline_command == ''
    assert statement.raw == line
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == ''
    assert statement.output_to == ''


def test_parse_command_only_strips_line(parser) -> None:
    line = '  help history  '
    statement = parser.parse_command_only(line)
    assert statement == 'history'
    assert statement.args == statement
    assert statement.arg_list == []
    assert statement.command == 'help'
    assert statement.command_and_args == line.strip()
    assert statement.multiline_command == ''
    assert statement.raw == line
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == ''
    assert statement.output_to == ''


def test_parse_command_only_expands_alias(parser) -> None:
    line = 'fake foobar.py "somebody.py'
    statement = parser.parse_command_only(line)
    assert statement == 'foobar.py "somebody.py'
    assert statement.args == statement
    assert statement.arg_list == []
    assert statement.command == 'run_pyscript'
    assert statement.command_and_args == 'run_pyscript foobar.py "somebody.py'
    assert statement.multiline_command == ''
    assert statement.raw == line
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == ''
    assert statement.output_to == ''


def test_parse_command_only_expands_shortcuts(parser) -> None:
    line = '!cat foobar.txt'
    statement = parser.parse_command_only(line)
    assert statement == 'cat foobar.txt'
    assert statement.args == statement
    assert statement.arg_list == []
    assert statement.command == 'shell'
    assert statement.command_and_args == 'shell cat foobar.txt'
    assert statement.multiline_command == ''
    assert statement.raw == line
    assert statement.multiline_command == ''
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == ''
    assert statement.output_to == ''


def test_parse_command_only_quoted_args(parser) -> None:
    line = 'l "/tmp/directory with spaces/doit.sh"'
    statement = parser.parse_command_only(line)
    assert statement == 'ls -al "/tmp/directory with spaces/doit.sh"'
    assert statement.args == statement
    assert statement.arg_list == []
    assert statement.command == 'shell'
    assert statement.command_and_args == line.replace('l', 'shell ls -al')
    assert statement.multiline_command == ''
    assert statement.raw == line
    assert statement.multiline_command == ''
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == ''
    assert statement.output_to == ''


def test_parse_command_only_unclosed_quote(parser) -> None:
    # Quoted trailing spaces will be preserved
    line = 'command with unclosed "quote     '
    statement = parser.parse_command_only(line)
    assert statement == 'with unclosed "quote     '
    assert statement.args == statement
    assert statement.arg_list == []
    assert statement.command == 'command'
    assert statement.command_and_args == line
    assert statement.multiline_command == ''
    assert statement.raw == line
    assert statement.multiline_command == ''
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == ''
    assert statement.output_to == ''


@pytest.mark.parametrize(
    ('line', 'args'),
    [
        ('helpalias > out.txt', '> out.txt'),
        ('helpalias>out.txt', '>out.txt'),
        ('helpalias >> out.txt', '>> out.txt'),
        ('helpalias>>out.txt', '>>out.txt'),
        ('help|less', '|less'),
        ('helpalias;', ';'),
        ('help ;;', ';;'),
        ('help; ;;', '; ;;'),
    ],
)
def test_parse_command_only_specialchars(parser, line, args) -> None:
    statement = parser.parse_command_only(line)
    assert statement == args
    assert statement.args == args
    assert statement.command == 'help'
    assert statement.multiline_command == ''
    assert statement.raw == line
    assert statement.multiline_command == ''
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == ''
    assert statement.output_to == ''


@pytest.mark.parametrize(
    'line',
    [
        '',
        ';',
        ';;',
        ';; ;',
        '&',
        '& &',
        ' && &',
        '>',
        "'",
        '"',
        '|',
    ],
)
def test_parse_command_only_empty(parser, line) -> None:
    statement = parser.parse_command_only(line)
    assert statement == ''
    assert statement.args == statement
    assert statement.arg_list == []
    assert statement.command == ''
    assert statement.command_and_args == ''
    assert statement.multiline_command == ''
    assert statement.raw == line
    assert statement.multiline_command == ''
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert statement.pipe_to == ''
    assert statement.output == ''
    assert statement.output_to == ''


def test_parse_command_only_multiline(parser) -> None:
    line = 'multiline with partially "open quotes and no terminator'
    statement = parser.parse_command_only(line)
    assert statement.command == 'multiline'
    assert statement.multiline_command == 'multiline'
    assert statement == 'with partially "open quotes and no terminator'
    assert statement.command_and_args == line
    assert statement.args == statement


def test_statement_initialization() -> None:
    string = 'alias'
    statement = cmd2.Statement(string)
    assert string == statement
    assert statement.args == statement
    assert statement.raw == ''
    assert statement.command == ''
    assert isinstance(statement.arg_list, list)
    assert not statement.arg_list
    assert isinstance(statement.argv, list)
    assert not statement.argv
    assert statement.multiline_command == ''
    assert statement.terminator == ''
    assert statement.suffix == ''
    assert isinstance(statement.pipe_to, str)
    assert not statement.pipe_to
    assert statement.output == ''
    assert statement.output_to == ''


def test_statement_is_immutable() -> None:
    string = 'foo'
    statement = cmd2.Statement(string)
    assert string == statement
    assert statement.args == statement
    assert statement.raw == ''
    with pytest.raises(dataclasses.FrozenInstanceError):
        statement.args = 'bar'
    with pytest.raises(dataclasses.FrozenInstanceError):
        statement.raw = 'baz'


def test_statement_as_dict(parser) -> None:
    # Make sure to_dict() results can be restored to identical Statement
    statement = parser.parse("!ls > out.txt")
    assert statement == Statement.from_dict(statement.to_dict())

    statement = parser.parse("!ls | grep text")
    assert statement == Statement.from_dict(statement.to_dict())

    statement = parser.parse("multiline arg; suffix")
    assert statement == Statement.from_dict(statement.to_dict())

    # from_dict() should raise KeyError if required field is missing
    statement = parser.parse("command")
    statement_dict = statement.to_dict()
    del statement_dict[Statement._args_field]

    with pytest.raises(KeyError):
        Statement.from_dict(statement_dict)


def test_is_valid_command_invalid(mocker, parser) -> None:
    # Non-string command
    valid, errmsg = parser.is_valid_command(5)
    assert not valid
    assert 'must be a string' in errmsg

    mock = mocker.MagicMock()
    valid, errmsg = parser.is_valid_command(mock)
    assert not valid
    assert 'must be a string' in errmsg

    # Empty command
    valid, errmsg = parser.is_valid_command('')
    assert not valid
    assert 'cannot be an empty string' in errmsg

    # Start with the comment character
    valid, errmsg = parser.is_valid_command(constants.COMMENT_CHAR)
    assert not valid
    assert 'cannot start with the comment character' in errmsg

    # Starts with shortcut
    valid, errmsg = parser.is_valid_command('!ls')
    assert not valid
    assert 'cannot start with a shortcut' in errmsg

    # Contains whitespace
    valid, errmsg = parser.is_valid_command('shell ls')
    assert not valid
    assert 'cannot contain: whitespace, quotes,' in errmsg

    # Contains a quote
    valid, errmsg = parser.is_valid_command('"shell"')
    assert not valid
    assert 'cannot contain: whitespace, quotes,' in errmsg

    # Contains a redirector
    valid, errmsg = parser.is_valid_command('>shell')
    assert not valid
    assert 'cannot contain: whitespace, quotes,' in errmsg

    # Contains a terminator
    valid, errmsg = parser.is_valid_command(';shell')
    assert not valid
    assert 'cannot contain: whitespace, quotes,' in errmsg


def test_is_valid_command_valid(parser) -> None:
    # Valid command
    valid, errmsg = parser.is_valid_command('shell')
    assert valid
    assert not errmsg

    # Subcommands can start with shortcut
    valid, errmsg = parser.is_valid_command('!subcmd', is_subcommand=True)
    assert valid
    assert not errmsg


def test_macro_normal_arg_pattern() -> None:
    # This pattern matches digits surrounded by exactly 1 brace on a side and 1 or more braces on the opposite side
    from cmd2.parsing import (
        MacroArg,
    )

    pattern = MacroArg.macro_normal_arg_pattern

    # Valid strings
    matches = pattern.findall('{5}')
    assert matches == ['{5}']

    matches = pattern.findall('{233}')
    assert matches == ['{233}']

    matches = pattern.findall('{{{{{4}')
    assert matches == ['{4}']

    matches = pattern.findall('{2}}}}}')
    assert matches == ['{2}']

    matches = pattern.findall('{3}{4}{5}')
    assert matches == ['{3}', '{4}', '{5}']

    matches = pattern.findall('{3} {4} {5}')
    assert matches == ['{3}', '{4}', '{5}']

    matches = pattern.findall('{3} {{{4} {5}}}}')
    assert matches == ['{3}', '{4}', '{5}']

    matches = pattern.findall('{3} text {4} stuff {5}}}}')
    assert matches == ['{3}', '{4}', '{5}']

    # Unicode digit
    matches = pattern.findall('{\N{ARABIC-INDIC DIGIT ONE}}')
    assert matches == ['{\N{ARABIC-INDIC DIGIT ONE}}']

    # Invalid strings
    matches = pattern.findall('5')
    assert not matches

    matches = pattern.findall('{5')
    assert not matches

    matches = pattern.findall('5}')
    assert not matches

    matches = pattern.findall('{{5}}')
    assert not matches

    matches = pattern.findall('{5text}')
    assert not matches


def test_macro_escaped_arg_pattern() -> None:
    # This pattern matches digits surrounded by 2 or more braces on both sides
    from cmd2.parsing import (
        MacroArg,
    )

    pattern = MacroArg.macro_escaped_arg_pattern

    # Valid strings
    matches = pattern.findall('{{5}}')
    assert matches == ['{{5}}']

    matches = pattern.findall('{{233}}')
    assert matches == ['{{233}}']

    matches = pattern.findall('{{{{{4}}')
    assert matches == ['{{4}}']

    matches = pattern.findall('{{2}}}}}')
    assert matches == ['{{2}}']

    matches = pattern.findall('{{3}}{{4}}{{5}}')
    assert matches == ['{{3}}', '{{4}}', '{{5}}']

    matches = pattern.findall('{{3}} {{4}} {{5}}')
    assert matches == ['{{3}}', '{{4}}', '{{5}}']

    matches = pattern.findall('{{3}} {{{4}} {{5}}}}')
    assert matches == ['{{3}}', '{{4}}', '{{5}}']

    matches = pattern.findall('{{3}} text {{4}} stuff {{5}}}}')
    assert matches == ['{{3}}', '{{4}}', '{{5}}']

    # Unicode digit
    matches = pattern.findall('{{\N{ARABIC-INDIC DIGIT ONE}}}')
    assert matches == ['{{\N{ARABIC-INDIC DIGIT ONE}}}']

    # Invalid strings
    matches = pattern.findall('5')
    assert not matches

    matches = pattern.findall('{{5')
    assert not matches

    matches = pattern.findall('5}}')
    assert not matches

    matches = pattern.findall('{5}')
    assert not matches

    matches = pattern.findall('{{5text}}')
    assert not matches



#!/usr/bin/env python
"""A simple example demonstrating how to integrate tab completion with argparse-based commands."""

import argparse

import rich.box
from rich.style import Style
from rich.table import Table
from rich.text import Text

from cmd2 import (
    Cmd,
    Cmd2ArgumentParser,
    Cmd2Style,
    Color,
    CompletionError,
    CompletionItem,
    with_argparser,
)

# Data source for argparse.choices
food_item_strs = ['Pizza', 'Ham', 'Ham Sandwich', 'Potato']


class ArgparseCompletion(Cmd):
    def __init__(self) -> None:
        super().__init__(include_ipy=True)
        self.sport_item_strs = ['Bat', 'Basket', 'Basketball', 'Football', 'Space Ball']

    def choices_provider(self) -> list[str]:
        """A choices provider is useful when the choice list is based on instance data of your application."""
        return self.sport_item_strs

    def choices_completion_error(self) -> list[str]:
        """CompletionErrors can be raised if an error occurs while tab completing.

        Example use cases
            - Reading a database to retrieve a tab completion data set failed
            - A previous command line argument that determines the data set being completed is invalid
        """
        if self.debug:
            return self.sport_item_strs
        raise CompletionError("debug must be true")

    def choices_completion_item(self) -> list[CompletionItem]:
        """Return CompletionItem instead of strings. These give more context to what's being tab completed."""
        fancy_item = Text.assemble(
            "These things can\ncontain newlines and\n",
            Text("styled text!!", style=Style(color=Color.BRIGHT_YELLOW, underline=True)),
        )

        table_item = Table(
            "Left Column",
            "Right Column",
            box=rich.box.ROUNDED,
            border_style=Cmd2Style.TABLE_BORDER,
        )
        table_item.add_row("Yes, it's true.", "CompletionItems can")
        table_item.add_row("even display description", "data in tables!")

        items = {
            1: "My item",
            2: "Another item",
            3: "Yet another item",
            4: fancy_item,
            5: table_item,
        }
        return [CompletionItem(item_id, [description]) for item_id, description in items.items()]

    def choices_arg_tokens(self, arg_tokens: dict[str, list[str]]) -> list[str]:
        """If a choices or completer function/method takes a value called arg_tokens, then it will be
        passed a dictionary that maps the command line tokens up through the one being completed
        to their argparse argument name.  All values of the arg_tokens dictionary are lists, even if
        a particular argument expects only 1 token.
        """
        # Check if choices_provider flag has appeared
        values = ['choices_provider', 'flag']
        if 'choices_provider' in arg_tokens:
            values.append('is {}'.format(arg_tokens['choices_provider'][0]))
        else:
            values.append('not supplied')
        return values

    # Parser for example command
    example_parser = Cmd2ArgumentParser(
        description="Command demonstrating tab completion with argparse\nNotice even the flags of this command tab complete"
    )

    # Tab complete from a list using argparse choices. Set metavar if you don't
    # want the entire choices list showing in the usage text for this command.
    example_parser.add_argument('--choices', choices=food_item_strs, metavar="CHOICE", help="tab complete using choices")

    # Tab complete from choices provided by a choices_provider
    example_parser.add_argument(
        '--choices_provider', choices_provider=choices_provider, help="tab complete using a choices_provider"
    )

    # Tab complete using a completer
    example_parser.add_argument('--completer', completer=Cmd.path_complete, help="tab complete using a completer")

    # Demonstrate raising a CompletionError while tab completing
    example_parser.add_argument(
        '--completion_error',
        choices_provider=choices_completion_error,
        help="raise a CompletionError while tab completing if debug is False",
    )

    # Demonstrate returning CompletionItems instead of strings
    example_parser.add_argument(
        '--completion_item',
        choices_provider=choices_completion_item,
        metavar="ITEM_ID",
        descriptive_headers=["Description"],
        help="demonstrate use of CompletionItems",
    )

    # Demonstrate use of arg_tokens dictionary
    example_parser.add_argument(
        '--arg_tokens', choices_provider=choices_arg_tokens, help="demonstrate use of arg_tokens dictionary"
    )

    @with_argparser(example_parser)
    def do_example(self, _: argparse.Namespace) -> None:
        """The example command."""
        self.poutput("I do nothing")


if __name__ == '__main__':
    import sys

    app = ArgparseCompletion()
    sys.exit(app.cmdloop())
